/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    navigation::ArmarXObjects::dynamic_scene_provider
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */


#pragma once


// #include <mutex>

#include <VirtualRobot/VirtualRobot.h>

#include "ArmarXCore/core/services/tasks/TaskUtil.h"
#include <ArmarXCore/core/Component.h>

#include "RobotAPI/libraries/armem/client/plugins/PluginUser.h"
#include "RobotAPI/libraries/armem/client/plugins/ReaderWriterPlugin.h"
#include "RobotAPI/libraries/armem_robot_state/client/common/VirtualRobotReader.h"
#include "RobotAPI/libraries/armem_vision/client/laser_scanner_features/Reader.h"

#include "VisionX/libraries/armem_human/client/HumanPoseReader.h"

// #include <ArmarXCore/libraries/ArmarXCoreComponentPlugins/DebugObserverComponentPlugin.h>

// #include <ArmarXGui/libraries/ArmarXGuiComponentPlugins/LightweightRemoteGuiComponentPlugin.h>

#include "RobotAPI/libraries/armem_vision/client/occupancy_grid/Reader.h"
#include <RobotAPI/libraries/ArmarXObjects/plugins/ObjectPoseClientPlugin.h>
#include <RobotAPI/libraries/RobotAPIComponentPlugins/ArVizComponentPlugin.h>

#include "armarx/navigation/components/dynamic_scene_provider/ArVizDrawer.h"
#include "armarx/navigation/human/HumanTracker.h"
#include "armarx/navigation/memory/client/costmap/Reader.h"
#include "armarx/navigation/memory/client/human/Writer.h"
#include <armarx/navigation/components/dynamic_scene_provider/ComponentInterface.h>


namespace armarx::navigation::components::dynamic_scene_provider
{

    class Component :
        virtual public armarx::Component,
        virtual public armarx::navigation::components::dynamic_scene_provider::ComponentInterface,
        // , virtual public armarx::DebugObserverComponentPluginUser
        // , virtual public armarx::LightweightRemoteGuiComponentPluginUser
        virtual public armarx::ArVizComponentPluginUser,
        virtual public armarx::armem::client::plugins::PluginUser,
        virtual public ObjectPoseClientPluginUser
    {
    public:
        Component();

        /// @see armarx::ManagedIceObject::getDefaultName()
        std::string getDefaultName() const override;

        /// Get the component's default name.
        static std::string GetDefaultName();


    protected:
        /// @see PropertyUser::createPropertyDefinitions()
        armarx::PropertyDefinitionsPtr createPropertyDefinitions() override;

        /// @see armarx::ManagedIceObject::onInitComponent()
        void onInitComponent() override;

        /// @see armarx::ManagedIceObject::onConnectComponent()
        void onConnectComponent() override;

        /// @see armarx::ManagedIceObject::onDisconnectComponent()
        void onDisconnectComponent() override;

        /// @see armarx::ManagedIceObject::onExitComponent()
        void onExitComponent() override;


        /* (Requires armarx::LightweightRemoteGuiComponentPluginUser.)
        /// This function should be called once in onConnect() or when you
        /// need to re-create the Remote GUI tab.
        void createRemoteGuiTab();

        /// After calling `RemoteGui_startRunningTask`, this function is
        /// called periodically in a separate thread. If you update variables,
        /// make sure to synchronize access to them.
        void RemoteGui_update() override;
        */


    private:
        // Private methods go here.

        // Forward declare `Properties` if you used it before its defined.
        // struct Properties;

        /* (Requires the armarx::ArVizComponentPluginUser.)
        /// Draw some boxes in ArViz.
        void drawBoxes(const Properties& p, viz::Client& arviz);
        */

        void runPeriodically();

        VirtualRobot::RobotPtr robot = nullptr;


    private:
        static const std::string defaultName;


        // Private member variables go here.


        /// Properties shown in the Scenario GUI.
        struct Properties
        {
            int taskPeriodMs = 100;

            struct
            {
                std::string providerName = "LaserScannerFeatureExtraction";
                std::string name = ""; // all
            } laserScannerFeatures;


            struct
            {
                std::string name = "Armar6";
            } robot;

            struct
            {
                std::string providerName = "CartographerMappingAndLocalization";
                std::string name = ""; // all

                float freespaceThreshold = 0.45F;
                float occupiedThreshold = 0.55;
            } occupancyGrid;

            std::string humanPoseProvider = "OpenNIPointCloudProvider";
        };
        Properties properties;
        /* Use a mutex if you access variables from different threads
         * (e.g. ice functions and RemoteGui_update()).
        std::mutex propertiesMutex;
        */


        /* (Requires the armarx::LightweightRemoteGuiComponentPluginUser.)
        /// Tab shown in the Remote GUI.
        struct RemoteGuiTab : armarx::RemoteGui::Client::Tab
        {
            armarx::RemoteGui::Client::LineEdit boxLayerName;
            armarx::RemoteGui::Client::IntSpinBox numBoxes;

            armarx::RemoteGui::Client::Button drawBoxes;
        };
        RemoteGuiTab tab;
        */


        /* (Requires the armarx::ArVizComponentPluginUser.)
         * When used from different threads, an ArViz client needs to be synchronized.
        /// Protects the arviz client inherited from the ArViz plugin.
        std::mutex arvizMutex;
        */

        std::optional<ArVizDrawer> arvizDrawer;


        PeriodicTask<Component>::pointer_type task;

        template <typename T>
        using ReaderWriterPlugin = armem::client::plugins::ReaderWriterPlugin<T>;

        ReaderWriterPlugin<armem::human::client::Reader>* humanPoseReaderPlugin = nullptr;

        ReaderWriterPlugin<armem::vision::laser_scanner_features::client::Reader>*
            laserScannerFeaturesReaderPlugin = nullptr;

        ReaderWriterPlugin<armem::robot_state::VirtualRobotReader>* virtualRobotReaderPlugin =
            nullptr;

        ReaderWriterPlugin<memory::client::costmap::Reader>* costmapReaderPlugin = nullptr;

        ReaderWriterPlugin<armem::vision::occupancy_grid::client::Reader>*
            occupancyGridReaderPlugin = nullptr;

        ReaderWriterPlugin<memory::client::human::Writer>* humanWriterPlugin = nullptr;


        human::HumanTracker humanTracker;
    };

} // namespace armarx::navigation::components::dynamic_scene_provider
