/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    navigation::ArmarXObjects::human_simulator
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */


#include "Component.h"

#include <cstdlib>

#include <VirtualRobot/Random.h>

#include "ArmarXCore/core/logging/Logging.h"
#include <ArmarXCore/libraries/DecoupledSingleComponent/Decoupled.h>

#include "armarx/navigation/algorithms/Costmap.h"
#include "armarx/navigation/core/basic_types.h"
#include "armarx/navigation/human/types.h"
#include "armarx/navigation/simulation/SimulatedHuman.h"

// Include headers you only need in function definitions in the .cpp.

// #include <Eigen/Core>

// #include <SimoxUtility/color/Color.h>


namespace armarx::navigation::components::human_simulator
{
    Component::Component()
    {
        addPlugin(virtualRobotReaderPlugin);
        addPlugin(costmapReaderPlugin);
        addPlugin(humanWriterPlugin);
    }


    const std::string Component::defaultName = "human_simulator";


    armarx::PropertyDefinitionsPtr
    Component::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr def =
            new armarx::ComponentPropertyDefinitions(getConfigIdentifier());

        // Publish to a topic (passing the TopicListenerPrx).
        // def->topic(myTopicListener);

        // Subscribe to a topic (passing the topic name).
        // def->topic<PlatformUnitListener>("MyTopic");

        // Use (and depend on) another component (passing the ComponentInterfacePrx).
        // def->component(myComponentProxy)


        // Add a required property. (The component won't start without a value being set.)
        // def->required(properties.boxLayerName, "p.box.LayerName", "Name of the box layer in ArViz.");

        // Add an optionalproperty.
        def->optional(properties.taskPeriodMs, "p.taskPeriodMs", "");
        def->optional(properties.nHumans, "p.nHumans", "Number of humans to spawn.");
        
        def->optional(properties.humanParams.goalDistanceThreshold, "p.human.goalDistanceThreshold", "");
        def->optional(properties.humanParams.minLinearVelocity, "p.human.maxLinearVelocity", "");
        def->optional(properties.humanParams.maxLinearVelocity, "p.human.maxLinearVelocity", "");

        return def;
    }


    void
    Component::onInitComponent()
    {
        // Topics and properties defined above are automagically registered.

        // Keep debug observer data until calling `sendDebugObserverBatch()`.
        // (Requies the armarx::DebugObserverComponentPluginUser.)
        // setDebugObserverBatchModeEnabled(true);
    }


    void
    Component::onConnectComponent()
    {
        // Do things after connecting to topics and components.

        /* (Requies the armarx::DebugObserverComponentPluginUser.)
        // Use the debug observer to log data over time.
        // The data can be viewed in the ObserverView and the LivePlotter.
        // (Before starting any threads, we don't need to lock mutexes.)
        {
            setDebugObserverDatafield("numBoxes", properties.numBoxes);
            setDebugObserverDatafield("boxLayerName", properties.boxLayerName);
            sendDebugObserverBatch();
        }
        */

        /* (Requires the armarx::ArVizComponentPluginUser.)
        // Draw boxes in ArViz.
        // (Before starting any threads, we don't need to lock mutexes.)
        drawBoxes(properties, arviz);
        */

        /* (Requires the armarx::LightweightRemoteGuiComponentPluginUser.)
        // Setup the remote GUI.
        {
            createRemoteGuiTab();
            RemoteGui_startRunningTask();
        }
        */


        ///

        //
        //  Costmaps
        //

        ARMARX_INFO << "Querying costmap";

        const algorithms::Costmap costmap = [&]() -> algorithms::Costmap
        {
            while (true)
            {
                const auto timestamp = Clock::Now();

                const memory::client::costmap::Reader::Query costmapQuery{
                    .providerName = "distance_to_obstacle_costmap_provider", // TODO check
                    .name = "distance_to_obstacles",
                    .timestamp = timestamp};

                const memory::client::costmap::Reader::Result costmapResult =
                    costmapReaderPlugin->get().query(costmapQuery);

                // if costmap is not available yet, wait
                if (costmapResult.status != memory::client::costmap::Reader::Result::Success)
                {
                    Clock::WaitFor(Duration::MilliSeconds(100));
                    continue;
                }

                ARMARX_CHECK_EQUAL(costmapResult.status,
                                   memory::client::costmap::Reader::Result::Success);

                ARMARX_TRACE;
                ARMARX_CHECK(costmapResult.costmap.has_value());
                return costmapResult.costmap.value();
            };
        }();

        /// create humans

        simulatedHumans.clear();

        for (std::size_t i = 0; i < properties.nHumans; i++)
        {
            simulatedHumans.emplace_back(costmap);
        }


        ///

        task = new PeriodicTask<Component>(
            this, &Component::runPeriodically, properties.taskPeriodMs, false, "runningTask");
        task->start();
    }


    void
    Component::onDisconnectComponent()
    {
        task->stop();
    }


    void
    Component::onExitComponent()
    {
    }


    std::string
    Component::getDefaultName() const
    {
        return Component::defaultName;
    }


    std::string
    Component::GetDefaultName()
    {
        return Component::defaultName;
    }

    void
    Component::runPeriodically()
    {
        // obtain data from perception

        const DateTime timestamp = Clock::Now();

        //
        // Robot
        //

        // ARMARX_CHECK(virtualRobotReaderPlugin->get().synchronizeRobot(*robot, timestamp));
        // const core::Pose global_T_robot(robot->getGlobalPose());

        // ARMARX_INFO << "Robot position: " << global_T_robot.translation().head<2>();


        human::Humans humans;

        for (auto& simulatedHuman : simulatedHumans)
        {
            humans.push_back(simulatedHuman.update());
        }

        ARMARX_VERBOSE << "Updating humans.";
        humanWriterPlugin->get().store(humans, getName(), timestamp);
    }


    /* (Requires the armarx::LightweightRemoteGuiComponentPluginUser.)
    void
    Component::createRemoteGuiTab()
    {
        using namespace armarx::RemoteGui::Client;

        // Setup the widgets.

        tab.boxLayerName.setValue(properties.boxLayerName);

        tab.numBoxes.setValue(properties.numBoxes);
        tab.numBoxes.setRange(0, 100);

        tab.drawBoxes.setLabel("Draw Boxes");

        // Setup the layout.

        GridLayout grid;
        int row = 0;
        {
            grid.add(Label("Box Layer"), {row, 0}).add(tab.boxLayerName, {row, 1});
            ++row;

            grid.add(Label("Num Boxes"), {row, 0}).add(tab.numBoxes, {row, 1});
            ++row;

            grid.add(tab.drawBoxes, {row, 0}, {2, 1});
            ++row;
        }

        VBoxLayout root = {grid, VSpacer()};
        RemoteGui_createTab(getName(), root, &tab);
    }


    void
    Component::RemoteGui_update()
    {
        if (tab.boxLayerName.hasValueChanged() || tab.numBoxes.hasValueChanged())
        {
            std::scoped_lock lock(propertiesMutex);
            properties.boxLayerName = tab.boxLayerName.getValue();
            properties.numBoxes = tab.numBoxes.getValue();

            {
                setDebugObserverDatafield("numBoxes", properties.numBoxes);
                setDebugObserverDatafield("boxLayerName", properties.boxLayerName);
                sendDebugObserverBatch();
            }
        }
        if (tab.drawBoxes.wasClicked())
        {
            // Lock shared variables in methods running in seperate threads
            // and pass them to functions. This way, the called functions do
            // not need to think about locking.
            std::scoped_lock lock(propertiesMutex, arvizMutex);
            drawBoxes(properties, arviz);
        }
    }
    */


    /* (Requires the armarx::ArVizComponentPluginUser.)
    void
    Component::drawBoxes(const Component::Properties& p, viz::Client& arviz)
    {
        // Draw something in ArViz (requires the armarx::ArVizComponentPluginUser.
        // See the ArVizExample in RobotAPI for more examples.

        viz::Layer layer = arviz.layer(p.boxLayerName);
        for (int i = 0; i < p.numBoxes; ++i)
        {
            layer.add(viz::Box("box_" + std::to_string(i))
                      .position(Eigen::Vector3f(i * 100, 0, 0))
                      .size(20).color(simox::Color::blue()));
        }
        arviz.commit(layer);
    }
    */


    ARMARX_REGISTER_COMPONENT_EXECUTABLE(Component, Component::GetDefaultName());

} // namespace armarx::navigation::components::human_simulator
