#include "HumanTracker.h"

#include "ArmarXCore/core/exceptions/local/ExpressionException.h"

#include "armarx/navigation/human/types.h"
#include <armarx/navigation/conversions/eigen.h>
#include <range/v3/range/conversion.hpp>
#include <range/v3/view/transform.hpp>

namespace armarx::navigation::human
{
    void
    HumanTracker::update(const CameraMeasurement& measurements)
    {
        prepareTrackedHumans(measurements.detectionTime);

        // calculate the poses according to the new received measurements
        std::vector<DetectedHuman> newPoses =
            measurements.humanPoses |
            ranges::views::transform(
                [measurements, this](const armem::human::HumanPose& humanPose) -> DetectedHuman {
                    return convertHumanPoseToDetectedHuman(measurements.detectionTime, humanPose);
                }) |
            ranges::to_vector;

        // associate the new poses from the measurement with the tracked humans
        associateHumans(newPoses);

        // add all not associated poses as new tracked humans
        for (const auto& detectedHuman : newPoses)
        {
            if (not detectedHuman.associated)
            {
                //add new tracked human to list of tracked humans
                trackedHumans.push_back(TrackedHuman{
                    .humanFilter = HumanFilter{detectedHuman.pose, detectedHuman.detectionTime},
                    .trackingId = detectedHuman.trackingId,
                    .associatedCluster = nullptr,
                    .associated = true});
            }
        }
    }

    std::vector<Cluster>
    HumanTracker::update(const LaserMeasurement& measurements)
    {
        // TODO do we receive cluster with correct threshold / expected size / as accurate as that
        // they distinguish multiple footprints?

        prepareTrackedHumans(measurements.detectionTime);

        // only look at the clusters that have a size reasonable enough to contain at most two
        // footprints
        std::vector<AdvancedCluster> potentialFootprints;
        std::vector<Cluster> unusedClusters;
        for (auto it = measurements.clusters.begin(); it != measurements.clusters.end();)
        {
            auto& cluster = *it;

            if (getClusterSize(cluster) <= 2 * parameters.maxFootprintSize)
            {
                //calculate center of cluster
                Eigen::Vector2f clusterCenter;
                for (auto& point : cluster.points)
                {
                    clusterCenter += point;
                }
                clusterCenter /= cluster.points.size();

                potentialFootprints.push_back(AdvancedCluster{
                    .center = clusterCenter, .cluster = cluster, .associated = false});
            }
            else
            {
                unusedClusters.push_back(cluster);
            }
        }


        // associate humans and clusters by their distances
        const auto sortedDistances = getSortedDistances(trackedHumans, potentialFootprints);

        for (auto& posDistance : sortedDistances)
        {
            if (posDistance.distance > parameters.maxAssociationDistance)
            {
                break;
            }
            if (posDistance.oldHuman->associated || posDistance.newCluster->associated)
            {
                continue;
            }
            // no footprint was found before, so associate first footprint
            if (!posDistance.oldHuman->associatedCluster)
            {
                posDistance.oldHuman->associatedCluster = posDistance.newCluster;
                posDistance.newCluster->associated = true;
            }
            // found both footprints, so associate both footprints and update human with new pose
            else
            {
                //calculate center of the points of both footprints combined
                auto& footprint1 = posDistance.oldHuman->associatedCluster->cluster.points;
                auto& footprint2 = posDistance.newCluster->cluster.points;

                Eigen::Vector2f centerPos;
                for (auto& point : footprint1)
                {
                    centerPos += point;
                }
                for (auto& point : footprint2)
                {
                    centerPos += point;
                }
                centerPos /= (footprint1.size() + footprint2.size());

                //create pose with orientation from old human
                core::Pose2D pose = core::Pose2D::Identity();
                pose.translation() = conv::to2D(centerPos);
                pose.linear() = posDistance.oldHuman->humanFilter.get().pose.linear();


                posDistance.oldHuman->humanFilter.update(pose, measurements.detectionTime);

                posDistance.oldHuman->associatedCluster = nullptr;
                posDistance.oldHuman->associated = true;
                posDistance.newCluster->associated = true;
            }
        }

        //associate humans where only one footprint was found
        for (auto& human : trackedHumans)
        {
            if (human.associated)
            {
                continue;
            }

            if (human.associatedCluster)
            {
                auto& footprint = human.associatedCluster;
                //create pose with orientation from old human
                core::Pose2D pose = core::Pose2D::Identity();
                pose.translation() = conv::to2D(footprint->center);
                pose.linear() = human.humanFilter.get().pose.linear();

                human.humanFilter.update(pose, measurements.detectionTime);

                human.associatedCluster = nullptr;
                human.associated = true;
            }
        }

        // add unused clusters to return list
        for (auto& cluster : potentialFootprints)
        {
            if (!cluster.associated)
            {
                unusedClusters.push_back(cluster.cluster);
            }
        }

        return unusedClusters;
    }

    std::vector<human::Human>
    HumanTracker::getTrackedHumans() const
    {
        return trackedHumans |
               ranges::views::transform([](const TrackedHuman& h) -> human::Human
                                        { return h.humanFilter.get(); }) |
               ranges::to_vector;
    }

    void
    HumanTracker::reset()
    {
        trackedHumans.clear();
    }


    HumanTracker::DetectedHuman
    HumanTracker::convertHumanPoseToDetectedHuman(const DateTime& time,
                                                  const armem::human::HumanPose& humanPose)
    {
        const std::map<std::string, armem::human::PoseKeypoint>& keypoints = humanPose.keypoints;
        ARMARX_CHECK_NOT_EMPTY(keypoints);

        // calculate the arithmetic mean of all keypoint positions
        Eigen::Vector3f centerPos;
        int size = 0;
        for (const auto& [_, v] : keypoints)
        {
            if (v.positionGlobal.has_value())
            {
                centerPos += v.positionGlobal.value().toEigen();
                size++;
            }
        }
        centerPos /= size;

        // calculate the yaw of the specified keypoint representing the orientation if it exists
        double yaw = 0;
        if (humanPose.keypoints.count(parameters.rotationKeypoint) > 0)
        {
            const armem::human::PoseKeypoint& refKeypoint =
                humanPose.keypoints.at(parameters.rotationKeypoint);
            ARMARX_CHECK(refKeypoint.orientationGlobal.has_value());
            Eigen::Quaternionf qhead = refKeypoint.orientationGlobal->toEigenQuaternion();
            Eigen::Vector3f vec(1, 0, 0);
            Eigen::Vector3f rot3 = qhead._transformVector(vec);
            Eigen::Vector2f rot2(rot3.x(), rot3.y());
            if (rot2.norm() != 0)
            {
                // calculate angle of rot2
                yaw = atan2(rot2.y(), rot2.x());
            }
            //old version using euler angles:
            //yaw = humanPose //from all human pose keypoints
            //          .keypoints
            //          .at(parameters.rotationKeypoint) //find the keypoint representing the head
            //          .orientationGlobal //get its global orientation
            //          ->toEigen()
            //          .eulerAngles(2, 1, 0)[0]; //and extract the yaw (rotation around z axis in
            //                                                                  global coordinates)
        }

        // create the new pose with the calculated position and yaw
        core::Pose2D pose = core::Pose2D::Identity();
        pose.translation() = conv::to2D(centerPos);
        pose.linear() = Eigen::Rotation2Df(yaw).toRotationMatrix();

        return {pose, humanPose.humanTrackingId, time, false};
    }

    std::vector<HumanTracker::PosHumanDistance>
    HumanTracker::getSortedDistances(std::vector<HumanTracker::TrackedHuman>& oldHumans,
                                     std::vector<HumanTracker::DetectedHuman>& newHumans)
    {
        std::vector<PosHumanDistance> posDistances;

        for (auto& oldHuman : oldHumans)
        {
            if (oldHuman.associated)
            {
                continue;
            }
            for (auto& newHuman : newHumans)
            {
                if (newHuman.associated)
                {
                    continue;
                }
                // calculate distance between every possible combination of tracked and detected
                // humans where none of them was associated
                posDistances.push_back(
                    {&oldHuman,
                     &newHuman,
                     (newHuman.pose.translation() - oldHuman.humanFilter.get().pose.translation())
                         .norm()});
            }
        }

        // sort the distances ascending by their numeric value
        std::sort(posDistances.begin(),
                  posDistances.end(),
                  [](const PosHumanDistance& a, const PosHumanDistance& b) -> bool
                  { return a.distance < b.distance; });

        return posDistances;
    }

    std::vector<HumanTracker::PosLaserDistance>
    HumanTracker::getSortedDistances(std::vector<TrackedHuman>& oldHumans,
                                     std::vector<AdvancedCluster>& newClusters)
    {
        std::vector<PosLaserDistance> posDistances;

        for (auto& oldHuman : oldHumans)
        {
            for (auto& newCluster : newClusters)
            {
                // calculate distance between every possible combination of tracked human and cluster
                posDistances.push_back(
                    {&oldHuman,
                     &newCluster,
                     (newCluster.center - oldHuman.humanFilter.get().pose.translation()).norm()});
            }
        }

        // sort the distances ascending by their numeric value
        std::sort(posDistances.begin(),
                  posDistances.end(),
                  [](const PosHumanDistance& a, const PosHumanDistance& b) -> bool
                  { return a.distance < b.distance; });

        return posDistances;
    }

    void
    HumanTracker::associateHumans(std::vector<DetectedHuman>& detectedHumans)
    {
        // first, associate humans by their tracking id
        for (auto& oldHuman : trackedHumans)
        {
            if (oldHuman.associated || not oldHuman.trackingId)
            {
                continue;
            }
            for (auto& newHuman : detectedHumans)
            {
                if (newHuman.associated || not newHuman.trackingId)
                {
                    continue;
                }
                // check for every possible combination of old and new humans (that were not already
                // associated and have a tracking id) if their tracking id is the same
                if (oldHuman.trackingId.value() == newHuman.trackingId.value())
                {
                    associate(&oldHuman, &newHuman);
                }
            }
        }

        // second, associate leftover humans by their distances
        const auto sortedDistances = getSortedDistances(trackedHumans, detectedHumans);

        for (auto& posDistance : sortedDistances)
        {
            if (posDistance.distance > parameters.maxAssociationDistance)
            {
                break;
            }
            if (posDistance.oldHuman->associated || posDistance.newHuman->associated)
            {
                continue;
            }
            // associate the pair with the currently smallest distance between non-associated humans
            associate(posDistance.oldHuman, posDistance.newHuman);
        }
    }

    void
    HumanTracker::associate(TrackedHuman* trackedHuman, DetectedHuman* detectedHuman)
    {
        ARMARX_CHECK(not trackedHuman->associated);
        ARMARX_CHECK(not detectedHuman->associated);

        trackedHuman->associated = true;
        detectedHuman->associated = true;

        trackedHuman->humanFilter.update(detectedHuman->pose, detectedHuman->detectionTime);
        trackedHuman->trackingId = detectedHuman->trackingId;
    }

    float
    HumanTracker::getClusterSize(Cluster cluster)
    {
        float maxSize = 0;
        for (auto it1 = cluster.points.begin(); it1 != cluster.points.end();)
        {
            auto& point1 = *it1;
            for (auto it2 = it1 + 1; it2 != cluster.points.end();)
            {
                auto& point2 = *it2;
                //TODO does subtraction work for vectors like this?
                maxSize = std::max(maxSize, (point2 - point1).norm());
            }
        }
        return maxSize;
    }

    void
    HumanTracker::prepareTrackedHumans(DateTime time)
    {
        // iterate over all existing tracked humans
        for (auto it = trackedHumans.begin(); it != trackedHumans.end();)
        {
            auto& human = *it;
            // when the tracked human recieved no new measurement for too long, remove it
            if ((time - human.humanFilter.get().detectionTime) >= parameters.maxTrackingAge)
            {
                it = trackedHumans.erase(it);
            }
            // otherwise the tracked human is prepared for association at the current point in time
            else
            {
                human.associatedCluster = nullptr;
                human.associated = false;
                human.humanFilter.propagation(time);
                it++;
            }
        }
    }

} // namespace armarx::navigation::human
